package project3

import org.scalatest._
import java.io.{ByteArrayOutputStream, PrintWriter}

// Define the stream method
trait TestOutput {
  import Language._

  val out = new ByteArrayOutputStream
  val pOut = new PrintWriter(out, true)
  def stream = pOut
  def emitCode(ast: Exp): Unit

  def code(ast: Exp) = {
    emitCode(ast)
    out.toString.stripLineEnd
  }
}

class CompilerTest extends TimedSuite {
  import Language._

  def runner(src: String) = new ASMRunner(src)

  def testCompiler(ast: Exp, res: Int) = {
    val interpreter = new X86Compiler with TestOutput

    val code = interpreter.code(ast)
    val asm = runner(code)

    assert(asm.assemble == 0, "Code generated couldn't be assembled")
    assert(asm.run == res, "Invalid result")
  }

  test("arithm") {
    testCompiler(LetRec(Nil, Lit(-21)), -21)
    testCompiler(LetRec(Nil, Prim("-", List(Lit(10), Lit(2)))), 8)
  }

  test("extra_credit_fn_as_var") {
    testCompiler(LetRec(Nil,Let("x",UnknownType,Ref("putchar"),Let("x$1",UnknownType,App(Ref("x"),List(Lit(97))),Lit(69)))), 69)
  }

  test("extra_credit_char_parse_toInt_toChar") {
    testCompiler(LetRec(Nil,Let("x$1",UnitType,App(Ref("putchar"),List(App(Ref("toInt"),List(Lit('s'))))),Let("x$2",UnitType,App(Ref("putchar"),List(App(Ref("toInt"),List(Lit('s'))))),Let("x$3",UnitType,App(Ref("putchar"),List(App(Ref("toInt"),List(Lit('i'))))),Lit(96))))), 96)
    testCompiler(LetRec(List(),App(Ref("toInt"),List(App(Ref("toChar"),List(Lit(420)))))), 164) // Int->Char narrow
  }

  test("func_decl_call") {
    val pos = Position(0,0,0,0,0,0)
    //def f(x:Int, y:Int, z:Char) = (x - y)*20 + 10 * toInt(z); f(6, 9, '0') == 420
    testCompiler(LetRec(List(FunDef("f",List(Arg("x",IntType,pos), Arg("y",IntType,Position(0,0,0,0,0,0)), Arg("z",CharType,Position(0,0,0,0,0,0))),IntType,Prim("+",List(Prim("*",List(Prim("-",List(Ref("x"), Ref("y"))), Lit(20))), Prim("*",List(Lit(10), App(Ref("toInt"),List(Ref("z"))))))))),App(Ref("f"),List(Lit(6), Lit(9), Lit('0')))), 420)
    //def f(x:Int) = 1 + x; val y = f(1); val z = f(2); y + z
    testCompiler(LetRec(List(FunDef("f",List(Arg("x",IntType,pos)),IntType,Prim("+",List(Lit(1), Ref("x"))))),Let("y",IntType,App(Ref("f"),List(Lit(1))),Let("z",IntType,App(Ref("f"),List(Lit(2))),Prim("+",List(Ref("y"), Ref("z")))))), 5)
  }
}
